import datetime
import sys
import traceback


__all__ = ['ScheduleError', 'print_exception', 'get_traceback',
           'is_ignored', 'ignore_exception', 'stop_ignore_exception']


# ========== Exception Handling ==========
IGNORE_PRINT_EXCEPTION = []


class ScheduleError(Exception):
    pass


def print_exception(exc, msg=None, error_cls=ScheduleError):
    """Print the given exception. If a message is given it will be prepended to the exception message with a \n.

    Args:
        exc (Exception): Exception that was raised.
        msg (str)[None]: Additional message to prepend to the exception.
        error_cls (Exception)[ServiceError]: New Exception class to print the exception as.
    """
    # Prepend the message to the exception if given
    if msg:
        msg = "\n".join((msg, str(exc)))
    else:
        msg = str(exc)

    if not is_ignored(msg, exc):
        exc_tb = get_traceback(exc)
        traceback.print_exception(error_cls, error_cls(msg), exc_tb)


def get_traceback(exc=None):
    """Get the exception traceback or the system traceback."""
    _, _, sys_tb = sys.exc_info()
    try:
        return exc.__traceback__
    except (AttributeError, Exception):
        return sys_tb


def is_ignored(msg=None, exc=None):
    """Return if a message should be ignored.

    Args:
        msg (str)[None]: String message to check if ignored
        exc (Exception)[None]: Exception to check if ignored (also checks if exception class is ignored).
    """
    # Get the exception class to check the ignore
    exc_cls = None
    if isinstance(exc, BaseException):
        exc_cls = type(exc)

    ignore_msg = msg and msg in IGNORE_PRINT_EXCEPTION
    ignore_exc = exc is not None and exc in IGNORE_PRINT_EXCEPTION
    ignore_exc_cls = exc_cls is not None and exc_cls in IGNORE_PRINT_EXCEPTION
    return ignore_msg or ignore_exc or ignore_exc_cls


def ignore_exception(exc):
    """Ignore the given exception, exception message."""
    if exc not in IGNORE_PRINT_EXCEPTION:
        IGNORE_PRINT_EXCEPTION.append(exc)


def stop_ignore_exception(exc, msg=None):
    """Stop ignoring an exception.

    Args:
        exc (Exception/str): Exception to stop ignoring
        msg (str)[None]: Message text to stop ignoring
    """
    if msg:
        exc = "\n".join((msg, str(exc)))

    try:
        IGNORE_PRINT_EXCEPTION.remove(exc)
    except (AttributeError, ValueError, TypeError, Exception):
        pass
    try:
        IGNORE_PRINT_EXCEPTION.remove(msg)
    except (AttributeError, ValueError, TypeError, Exception):
        pass



import os
import sys
import asyncio
from typing import Callable, Awaitable, Union

from serial_json import DataClass, loads, dumps

from ..utils import print_exception
from ..schedule import Schedule
from .messages import Message, Error, Quit, Update, RunCommand, ScheduleCommand, RunningSchedule, \
    ListSchedules, StopSchedule


__all__ = ['get_server', 'set_server', 'Scheduler']


SERVER = None


def get_server():
    global SERVER
    return SERVER


def set_server(value):
    global SERVER
    SERVER = value


class Scheduler(object):

    READ_SIZE = 4096

    def __init__(self, command_path=None, loop: asyncio.AbstractEventLoop = None):
        self._loop = loop

        self.command_path = command_path
        self.tasks = {}
        self.callbacks = {}
        self.server = None
        self.server_task = None

    def update_commands(self):
        """Read all of the files in the command path and register those commands to be able to run."""
        if self.command_path is None:
            return

        if self.command_path not in sys.path:
            sys.path.insert(0, self.command_path)

        for filename in os.listdir(self.command_path):
            try:
                name = os.path.splitext(filename)[0]
                mod = __import__(name)  # Use get_server() to register the callback.
            except (ImportError, Exception) as err:
                print_exception(err, msg='Could not import {}'.format(filename))

    @property
    def loop(self) -> 'asyncio.AbstractEventLoop':
        if self._loop is not None:
            return self._loop
        try:
            return asyncio.get_running_loop()
        except (RuntimeError, Exception):
            return asyncio.get_event_loop()

    @loop.setter
    def loop(self, loop: asyncio.AbstractEventLoop):
        self._loop = loop

    def register_callback(self, name: str = None, func: Callable[..., Awaitable[None]] = None):
        """Register a callback function to be callable from a received message.

        Args:
            name (str)[None]: Name of the callback function. If func the name will be populated from the func.__name__.
            func (callable)[None]: Callback function to call. If None a decorator wrapper will be returned.

        Returns:
            func (callable): If given func is None a decorator function will be returned else the given function.
        """
        if callable(name) and func is None:
            func = name
            name = None

        if func is None:
            def decorator(func):
                return self.register_callback(name, func)
            return decorator

        if name is None:
            name = func.__name__

        self.callbacks[name] = func
        return func

    async def handle_client(self, reader, writer):
        while self.is_serving() and not reader.at_eof() and not writer.is_closing():
            try:
                data = await reader.read(self.READ_SIZE)
                if not data:
                    continue
            except (TypeError, ValueError, Exception):
                break

            message = DataClass.from_json(data)
            if isinstance(message, Quit):
                writer.write(Message(message='Stopping server').json().encode())
                await writer.drain()
                await self.stop_async()

            elif isinstance(message, Update):
                self.update_commands()
                writer.write(Message(message='Updated commands').json().encode())
                await writer.drain()

            elif isinstance(message, ListSchedules):
                try:
                    li = ListSchedules(schedules=[RunningSchedule(name=name, schedule=Schedule(**item[1].dict()))
                                                  for name, item in self.tasks.items()])
                    writer.write(li.json().encode())
                except Exception as err:
                    print_exception(err, msg='Cannot read the list of schedules!')
                    writer.write(Error(message='Cannot read the list of schedules!').json().encode())
                await writer.drain()

            elif isinstance(message, RunCommand):
                try:
                    cmd = self.callbacks[message.callback_name]
                    cmd(*message.args, **message.kwargs)
                    writer.write(Message(message='Command "{}" ran successfully!'.format(message.callback_name)).json().encode())
                except Exception as err:
                    print_exception(err, msg='Could not run command "{}"'.format(message.callback_name))
                    writer.write(Error(message='Error in command "{}"'.format(message.callback_name)).json().encode())
                await writer.drain()

            elif isinstance(message, ScheduleCommand):
                try:
                    s = message.schedule
                    cmd = self.callbacks[message.callback_name]
                    self.add(message.name, s, cmd, *message.args, **message.kwargs)
                    writer.write(Message(message='Scheduled Command "{}" is running!'.format(message.callback_name)).json().encode())
                except Exception as err:
                    print_exception(err, msg='Could not run command "{}"'.format(message.callback_name))
                    writer.write(Error(message='Error in command "{}"'.format(message.callback_name)).json().encode())
                await writer.drain()

            elif isinstance(message, StopSchedule):
                try:
                    self.remove(message.name)
                    writer.write(Message(message='Stopped running the schedule named "{}"!'.format(message.name)).json().encode())
                except Exception as err:
                    print_exception(err, msg='Error while stopping schedule "{}"'.format(message.name))
                    writer.write(Error(message='Error while stopping schedule "{}"'.format(message.name)).json().encode())
                await writer.drain()

            else:
                writer.write(Error(message='Unknown command given!').json().encode())
                await writer.drain()

        # Close when ending
        writer.close()

    def is_serving(self) -> bool:
        """Return if the server is running."""
        try:
            return self.server.is_serving()
        except (AttributeError, Exception):
            return False

    def start(self, ip_addr: str = '127.0.0.1', port: int = 8000) -> 'Scheduler':
        """Add a task to start running the server forever."""
        self.server_task = self.loop.create_task(self.start_async(ip_addr, port), name='server')
        return self

    async def start_async(self, ip_addr: str = '127.0.0.1', port: int = 8000):
        """Start running the server forever."""
        self.server = await asyncio.start_server(self.handle_client, ip_addr, port)

        addr = self.server.sockets[0].getsockname()
        print(f'===== Serving on {addr} =====')

        try:
            async with self.server:
                await self.server.serve_forever()
        finally:
            print(f'===== Stopped serving on {addr} =====')

    def stop(self):
        """Stop running the server."""
        for k in self.server_task.all_tasks():
            print(k)
        print('========================')
        try:
            self.server.close()
        except (AttributeError, Exception):
            pass
        # try:
        #     self.server_task.cancel()
        # except (AttributeError, Exception):
        #     pass

        for k in self.server_task.all_tasks():
            print(k)

        return self

    async def stop_async(self):
        """Stop the server from running."""
        return self.stop()

    def add(self, name: str, schedule: Schedule, callback: Callable[..., Awaitable[None]] = None, *args, **kwargs):
        """Add a schedule to run.

        Args:
            name (str): Name of the schedule
            schedule (Schedule): Schedule to run.
            callback (callable/awaitable): Function to run on the given schedule.
            *args (tuple/object): Positional arguments to pass into the callback function.
            **kwargs (dict/object): Keyword arguments to pass into the callback function.
        """
        task = self.loop.create_task(schedule.run_async(callback, *args, **kwargs), name=name)
        if name in self.tasks:
            self.tasks[name].stop()
        self.tasks[name] = [task, schedule]

    def remove(self, name: str):
        """Remove and stop running a schedule.

        Args:
            name (str): Name of the schedule
        """
        if name in self.tasks:
            self.tasks[name][0].stop()

    # ========== Loop Functions ==========
    def create_task(self, coro, *, name=None):
        """Create a task to run on the loop"""
        return self.loop.create_task(coro, name=name)

    def run_until_complete(self, future):
        """Start running the asyncio event loop until the given future is complete."""
        return self.loop.run_until_complete(future)

    def run_forever(self):
        """Start running the asyncio event loop forever."""
        return self.loop.run_forever()



name = 'async_sched'
version = '1.0.5'
description = 'Use async to have functions run on a schedule.'
url = 'https://github.com/justengel/async_sched'
author = 'Justin Engel'
author_email = 'jtengel08@gmail.com'



import time
import inspect
import asyncio
import datetime
from typing import Union, Callable, Awaitable, Tuple, Optional, ClassVar

from serial_json import DataClass, field, field_property, MISSING, \
    Weekdays, weekdays_property, weekdays_attr_property, \
    datetime_property, time_property, timedelta_attr_property, seconds_property, make_datetime


__all__ = ['Schedule', 'RepeatSchedule']


class Schedule(DataClass):
    """Schedule a service to run.

    Example:

        Run async

        ..code-block:: python

            import asyncio
            import async_sched

            loop = asyncio.get_event_loop()

            async def print_sched(sched):
                print(sched)

            seconds = async_sched.Schedule(seconds=1)
            loop.create_task(seconds.run_async(print_sched, seconds))

            minutes = async_sched.Schedule(seconds=2)
            loop.create_task(minutes.run_async(print_sched, minutes))

            loop.run_forever()

        Run synchronous.

        ..code-block:: python

            def hello():
                print('hello')

            sched = Schedule(seconds=5)

            while True:
                sched.wait()
                hello()
                time.sleep(1)

    Args:
        days (int)[0]: Number of days when the next run should happen.
        hours (int)[0]: Number of hours when the next run should happen.
        minutes (int)[0]: Number of minutes when the next run should happen.
        seconds (int)[0]: Number of seconds when the next run should happen.
        milliseconds (int)[0]: Number of milliseconds when the next run should happen.
        microseconds (int)[0]: Number of microseconds when the next run should happen.
        weeks (int)[0]: Number of weeks for the interval
        interval (datetime.timedetla)[None]: Timedelta property for the time values.

        weekdays (Weekdays)[Weekdays()]: List of weekdays to run on.
        sundays (bool)[None]: Allow the schedule to run on sundays. If all weekdays are None allow all weekdays.
        mondays (bool)[None]: Allow the schedule to run on mondays. If all weekdays are None allow all weekdays.
        tuesdays (bool)[None]: Allow the schedule to run on tuesdays. If all weekdays are None allow all weekdays.
        wednesdays (bool)[None]: Allow the schedule to run on wednesdays. If all weekdays are None allow all weekdays.
        thursdays (bool)[None]: Allow the schedule to run on thursdays. If all weekdays are None allow all weekdays.
        fridays (bool)[None]: Allow the schedule to run on fridays. If all weekdays are None allow all weekdays.
        saturdays (bool)[None]: Allow the schedule to run on saturdays. If all weekdays are None allow all weekdays.

        repeat (bool)[True]: If True repeat the schedule else run once.
        at (Time/str)[None]: Time of day when the schedule should run.
        start_on (DateTime/str)[datetime.datetime.now()]: Date and time on which to start on.
        end_on (DateTime/str)[None]: Date and time on which to end on.
        last_run (DateTime/str)[None]: Date and time to make the next_run from.
        next_run (DateTime/str)[None]: Manually set the next run time.
    """
    days: int = field(0, skip_repr=0, skip_dict=0)
    hours: int = field(0, skip_repr=0, skip_dict=0)
    minutes: int = field(0, skip_repr=0, skip_dict=0)
    milliseconds: int = field(0, skip_repr=0, skip_dict=0)
    microseconds: int = field(0, skip_repr=0, skip_dict=0)
    seconds: Union[int, float] = seconds_property('seconds', skip_repr=0, skip_dict=0)
    weeks: int = field(0, skip_repr=0, skip_dict=0)
    interval: datetime.timedelta = timedelta_attr_property(required=False, repr=False, dict=False)

    weekdays: Weekdays = weekdays_property('weekdays', default_factory=Weekdays,
                                           skip_repr=Weekdays(), skip_dict=Weekdays())
    sunday: bool = weekdays_attr_property('weekdays', 'sunday', repr=False, dict=False)
    monday: bool = weekdays_attr_property('weekdays', 'monday', repr=False, dict=False)
    tuesday: bool = weekdays_attr_property('weekdays', 'tuesday', repr=False, dict=False)
    wednesday: bool = weekdays_attr_property('weekdays', 'wednesday', repr=False, dict=False)
    thursday: bool = weekdays_attr_property('weekdays', 'thursday', repr=False, dict=False)
    friday: bool = weekdays_attr_property('weekdays', 'friday', repr=False, dict=False)
    saturday: bool = weekdays_attr_property('weekdays', 'saturday', repr=False, dict=False)

    repeat: bool = False
    at: datetime.time = time_property('at', allow_none=True, required=False, skip_repr=None, skip_dict=None)
    start_on: datetime.datetime = datetime_property('start_on', default_factory=datetime.datetime.now, repr=False)
    end_on: datetime.datetime = datetime_property('end_on', allow_none=True, required=False, skip_repr=None, skip_dict=None)
    last_run: datetime.datetime = datetime_property('last_run', allow_none=True, required=False, repr=False, skip_dict=None)
    _next_run: Union[datetime.datetime, None] = field(default=None, repr=False, skip_dict=None)

    @field_property(default=None)
    def next_run(self) -> Union[datetime.datetime, None]:
        # Check end on
        now = datetime.datetime.now()
        if self.past_end(now):
            return None

        # Check for a set next run
        elif getattr(self, '_next_run', None) is not None:
            return self._next_run

        return self.create_run_time()

    @next_run.setter
    def next_run(self, value: Union[datetime.datetime, str, None]):
        if value is not None:
            value = make_datetime(value)
        self._next_run = value

    def run_in(self, now: datetime.datetime = None) -> Union[float, int]:
        """Return the number of seconds to wait until this should run."""
        if now is None:
            now = datetime.datetime.now()

        next_run = self.next_run
        if next_run is None:
            return -1
        elif now > next_run:
            return 0

        wait_time = (next_run - now).total_seconds()
        return wait_time

    def can_run(self, now: datetime.datetime = None) -> bool:
        """Return if this schedule can run now."""
        return self.run_in(now) == 0

    def past_end(self, now: datetime.datetime = None) -> bool:
        """Return if this datetime is past the end_on datetime and should stop running"""
        if now is None:
            now = datetime.datetime.now()
        return self.end_on and now >= self.end_on

    def wait(self, now: datetime.datetime = None) -> 'Schedule':
        """Wait until it is time to run."""
        time.sleep(self.run_in(now))
        return self

    async def wait_async(self, now: datetime.datetime = None) -> 'Schedule':
        """Asynchronous Wait until it is time to run."""
        await asyncio.sleep(self.run_in(now))
        return self

    def __await__(self) -> 'Schedule':
        yield from self.wait_async().__await__()
        return self

    def reschedule(self, now: datetime.datetime = None) -> 'Schedule':
        """Reset to get the next run time."""
        if now is None:
            now = datetime.datetime.now()

        # Setup the run times
        self.last_run = now
        self.next_run = None
        if not self.repeat:
            self.end_on = self.last_run
        return self

    async def reschedule_async(self, now: datetime.datetime = None) -> 'Schedule':
        """Reset to get the next run time."""
        return self.reschedule(now)

    def call(self, callback: Callable = None, *args, **kwargs) -> object:
        """Wait for the schedule and run the callback"""
        self.wait()
        self.reschedule()

        if callable(callback):
            return callback(*args, **kwargs)

    async def call_async(self, callback: Callable[..., Awaitable[None]] = None, *args, **kwargs) -> object:
        """Run the set callback and setup repeat if set."""
        await self.wait_async()
        await self.reschedule_async()
        if inspect.iscoroutinefunction(callback):
            return await callback(*args, **kwargs)
        elif callable(callback):
            return callback(*args, **kwargs)

    def run(self, callback: Callable = None, *args, **kwargs) -> 'Schedule':
        """Loop until and call this function until the schedule ends."""
        while not self.past_end():
            self.call(callback, *args, **kwargs)
        return self

    async def run_async(self, callback: Callable[..., Awaitable[None]] = None, *args, **kwargs) -> 'Schedule':
        """Generator to keep running this schedule repeatedly."""
        while not self.past_end():
            await self.call_async(callback, *args, **kwargs)
        return self

    WEEKDAY_ATTRS = tuple(Weekdays.DAYS)
    INTERVAL_ATTRS = ('weeks', 'days', 'hours', 'minutes', 'seconds', 'milliseconds')
    STRING_NAME_ATTRS = INTERVAL_ATTRS + ('repeat', 'at')

    def allowed_weekdays(self) -> Tuple[str]:
        """Return if all of the days of the week are None."""
        return tuple(self.weekdays)

    def is_allowed_weekday(self, dt: datetime.datetime) -> bool:
        """Return if the given datetime is on an allowed weekday."""
        return dt.weekday() in self.weekdays
        # return dt.strftime('%As').lower() in self.weekdays

    def make_at(self, dt: Union[datetime.datetime, datetime.timedelta]) -> datetime.datetime:
        """Make the given datetime run at the set "at" time if the "at" time was set."""
        if isinstance(dt, datetime.timedelta):
            today = datetime.datetime.today()
            return datetime.datetime(year=today.year, month=today.month, day=today.day,
                            hour=0, minute=0, second=0, microsecond=0) + dt
        else:
            return datetime.datetime(year=dt.year, month=dt.month, day=dt.day,
                            hour=self.at.hour, minute=self.at.minute, second=self.at.second,
                            microsecond=self.at.microsecond)

    def create_run_time(self) -> Union[datetime.datetime, None]:
        """Make the next_run datetime."""
        from_dt = self.last_run or self.start_on or datetime.datetime.now()
        dt = from_dt

        # Increment the interval and make at time
        dt = dt + self.interval
        if self.at is not None:
            dt = self.make_at(dt)

        # Check if allowed on this weekday
        i = 0
        while not self.is_allowed_weekday(dt):
            # Increment the interval and make at time
            dt = dt + datetime.timedelta(days=1)  # Increment by days
            if self.at is not None:
                dt = self.make_at(dt)

            i += 1
            if i > 7:
                self.end_on = from_dt  # No weekdays for this interval are allowed
                return None

        return dt

    # def __str__(self) -> str:
    #     attrs = tuple('{}={}'.format(attr, getattr(self, attr))
    #                   for attr in self.STRING_NAME_ATTRS if getattr(self, attr))
    #
    #     allowed_weekdays = self.allowed_weekdays()
    #     if allowed_weekdays != self.WEEKDAY_ATTRS:
    #         attrs = attrs + ('on=[' + ', '.join(allowed_weekdays) + ']',)
    #     return '{cls}({attrs})'.format(cls=self.__class__.__name__, attrs=', '.join(attrs))
    #
    # def __repr__(self) -> str:
    #     return '<{str} at {id}>'.format(str=self.__str__(), id=id(self))


class RepeatSchedule(Schedule):
    repeat: bool = True



import asyncio
from typing import Union

from serial_json import DataClass
from ..schedule import Schedule

from .messages import Quit, Update, RunCommand, ScheduleCommand, ListSchedules, StopSchedule


__all__ = ['Client']


class Client(object):

    READ_SIZE = 4096

    def __init__(self, ip_address: str = None, port: int = 8000, loop: asyncio.AbstractEventLoop = None):
        self._loop = loop

        self.reader = None
        self.writer = None
        self._is_connected = False

        self.ip_address = ip_address
        self.port = port

    @property
    def loop(self) -> 'asyncio.AbstractEventLoop':
        if self._loop is not None:
            return self._loop
        try:
            return asyncio.get_running_loop()
        except (RuntimeError, Exception):
            return asyncio.get_event_loop()

    @loop.setter
    def loop(self, loop: asyncio.AbstractEventLoop):
        self._loop = loop

    async def is_connected(self):
        """Return if the client is connected"""
        return self._is_connected

    def start(self, ip_addr: str = None, port: int = None) -> 'Client':
        """Add a task to start running the server forever."""
        self.server_task = self.loop.create_task(self.start_async(ip_addr, port), name='client')
        return self

    async def start_async(self, ip_address: str = None, port: int = None):
        """Start the connection."""
        if ip_address is not None:
            self.ip_address = ip_address
        if port is not None:
            self.port = port
        self.reader, self.writer = await asyncio.open_connection(self.ip_address, self.port)
        return self

    async def stop_async(self):
        """Stop the connection"""
        if self.writer is not None:
            self.writer.close()
            await self.writer.wait_closed()
        self.writer = None
        self.reader = None
        return self

    async def send_quit(self):
        """Send the quit command."""
        self.writer.write(Quit().json().encode())
        await self.writer.drain()

        data = await self.reader.read(self.READ_SIZE)
        message = DataClass.from_json(data)
        print(f'{message.message}')

    async def send_update(self):
        """Send the quit command."""
        self.writer.write(Update().json().encode())
        await self.writer.drain()

        data = await self.reader.read(self.READ_SIZE)
        message = DataClass.from_json(data)
        print(f'{message.message}')

    async def request_schedules(self):
        """Print the list of schedules."""
        self.writer.write(ListSchedules().json().encode())
        await self.writer.drain()

        data = await self.reader.read(self.READ_SIZE)
        if not data:
            data = await self.reader.read(self.READ_SIZE)
        message = DataClass.from_json(data)
        print('Running Schedules:')
        for running in message.schedules:
            print(f'{running.name} = {running.schedule}')

    async def run_command(self, callback_name, *args, **kwargs):
        """Print the list of schedules."""
        self.writer.write(RunCommand(callback_name=callback_name, args=args, kwargs=kwargs).json().encode())
        await self.writer.drain()

        data = await self.reader.read(self.READ_SIZE)
        message = DataClass.from_json(data)
        print(f'{message.message}')

    async def schedule_command(self, name: str, schedule: Schedule, callback_name, *args, **kwargs):
        """Print the list of schedules."""
        json = ScheduleCommand(name=name, schedule=schedule,
                               callback_name=callback_name, args=args, kwargs=kwargs).json()
        self.writer.write(json.encode())
        await self.writer.drain()

        data = await self.reader.read(self.READ_SIZE)
        message = DataClass.from_json(data)
        print(f'{message.message}')

    async def stop_schedule(self, name: str):
        """Stop a running schedule."""
        self.writer.write(StopSchedule(name=name).json().encode())
        await self.writer.drain()

        data = await self.reader.read(self.READ_SIZE)
        message = DataClass.from_json(data)
        print(f'{message.message}')

    async def __aenter__(self):
        if not await self.is_connected():
            await self.start_async()
        return self

    async def __aexit__(self, *args, **kwargs):
        await self.stop_async()

    def __enter__(self):
        if not self._is_connected:
            self.loop.run_until_complete(self.start_async())
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.loop.run_until_complete(self.stop_async())

        return exc_type is not None

    # ========== Loop Functions ==========
    def create_task(self, coro, *, name=None):
        """Create a task to run on the loop"""
        return self.loop.create_task(coro, name=name)

    def run_until_complete(self, future):
        """Start running the asyncio event loop until the given future is complete."""
        return self.loop.run_until_complete(future)

    def run_forever(self):
        """Start running the asyncio event loop forever."""
        return self.loop.run_forever()
